"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const R = __importStar(require("ramda"));
const Rx = __importStar(require("rxjs"));
const operators_1 = require("rxjs/operators");
const webSocket_1 = require("rxjs/webSocket");
const isomorphic_ws_1 = __importDefault(require("isomorphic-ws"));
const types_1 = require("@radarrelay/types");
const types_2 = require("../../types");
const debug = __importStar(require("../../debug"));
const common_1 = require("./common");
const cleanEvents_1 = require("../../utils/cleanEvents");
const subscribeMessage = (options, id) => {
    const baseToken = options.pair.base.symbol;
    const quoteToken = options.pair.quote.symbol;
    const message = {
        type: types_1.WebsocketRequestType.SUBSCRIBE,
        topic: types_1.WebsocketRequestTopic.BOOK,
        market: `${baseToken}-${quoteToken}`,
        requestId: id,
    };
    return message;
};
const unsubscribeMessage = (options) => {
    const baseToken = options.pair.base.symbol;
    const quoteToken = options.pair.quote.symbol;
    const message = {
        type: types_1.WebsocketRequestType.UNSUBSCRIBE,
        topic: types_1.WebsocketRequestTopic.BOOK,
        market: `${baseToken}-${quoteToken}`,
    };
    return message;
};
const normalizeNewOrderEvent = (options, event) => {
    const oid = Buffer.from(event.order.orderHash).toString('base64');
    return {
        id: oid,
        event: types_2.NormalizedMessageType.SET,
        exchange: types_2.Exchange.RADAR_RELAY,
        network: options.network,
        base: options.pair.base,
        quote: options.pair.quote,
        order: common_1.normalizeOrder(options, event.order),
    };
};
const normalizeFillOrderEvent = (options, event) => {
    const oid = Buffer.from(event.order.orderHash).toString('base64');
    return {
        id: oid,
        event: types_2.NormalizedMessageType.SET,
        exchange: types_2.Exchange.RADAR_RELAY,
        network: options.network,
        base: options.pair.base,
        quote: options.pair.quote,
        order: common_1.normalizeOrder(options, event.order),
    };
};
const normalizeCancelOrderEvent = (options, event) => ({
    event: types_2.NormalizedMessageType.REMOVE,
    exchange: types_2.Exchange.RADAR_RELAY,
    network: options.network,
    base: options.pair.base,
    quote: options.pair.quote,
    id: Buffer.from(event.orderHash).toString('base64'),
});
const normalizeRemoveOrderEvent = (options, event) => ({
    event: types_2.NormalizedMessageType.REMOVE,
    exchange: types_2.Exchange.RADAR_RELAY,
    network: options.network,
    base: options.pair.base,
    quote: options.pair.quote,
    id: Buffer.from(event.orderHash).toString('base64'),
});
const normalizeSnapshotEvent = (options, book) => ({
    event: types_2.NormalizedMessageType.SNAPSHOT,
    exchange: types_2.Exchange.RADAR_RELAY,
    network: options.network,
    base: options.pair.base,
    quote: options.pair.quote,
    orders: [].concat(book.asks.map(order => common_1.normalizeOrder(options, order)), book.bids.map(order => common_1.normalizeOrder(options, order))),
});
const isSnapshotEvent = R.allPass([R.has('asks'), R.has('bids')]);
const isOrderEvent = R.allPass([R.has('event'), R.has('action')]);
const isNewOrderEvent = R.allPass([
    isOrderEvent,
    R.propEq('action', 'NEW'),
]);
const isFillOrderEvent = R.allPass([
    isOrderEvent,
    R.propEq('action', 'FILL'),
]);
const isCancelOrderEvent = R.allPass([
    isOrderEvent,
    R.propEq('action', 'CANCEL'),
]);
const isRemoveOrderEvent = R.allPass([
    isOrderEvent,
    R.propEq('action', 'REMOVE'),
]);
const websocketConnections = {
    [types_2.Network.KOVAN]: undefined,
    [types_2.Network.MAINNET]: undefined,
};
const getWebsocketConnection = (options) => {
    if (typeof websocketConnections[options.network] !== 'undefined') {
        return websocketConnections[options.network];
    }
    const open$ = new Rx.Subject();
    const close$ = new Rx.Subject();
    const url = getWebsocketUrl(options);
    const connection$ = webSocket_1.webSocket({
        WebSocketCtor: isomorphic_ws_1.default,
        closeObserver: close$,
        openObserver: open$,
        url,
    });
    open$.subscribe(() => {
        debug.log('Opening connection.');
    });
    close$.subscribe(() => {
        debug.log('Closing connection.');
    });
    return (websocketConnections[options.network] = connection$);
};
let requestId = 0;
const getRequestId = () => {
    return (requestId = requestId + 1);
};
const getWebsocketUrl = (options) => {
    switch (options.network) {
        case types_2.Network.KOVAN:
            return 'wss://ws.kovan.radarrelay.com/v2';
        case types_2.Network.MAINNET:
            return 'wss://ws.radarrelay.com/v2';
        default:
            throw new Error('Invalid network.');
    }
};
exports.watch = (options) => {
    const ws$ = new Rx.Observable(observer => {
        const id = getRequestId();
        const ws$ = getWebsocketConnection(options);
        const socket$ = ws$.multiplex(() => subscribeMessage(options, id), () => unsubscribeMessage(options), R.propEq('requestId', id));
        return socket$.subscribe(observer);
    });
    const snapshot$ = ws$.pipe(operators_1.filter(R.propEq('type', types_1.WebsocketRequestType.SUBSCRIBE)), operators_1.tap(() => {
        const base = options.pair.base.symbol;
        const quote = options.pair.quote.symbol;
        debug.log(`Loading snapshot for market %s-%s.`, base, quote);
    }), operators_1.switchMap(() => common_1.fetchRadarBook(options)));
    // TODO: Make it so, that every time we are fetching a new snapshot (every
    // time we (re-)open the connection), the websocket messages are buffered
    // until the snapshot is emitted.
    const messages$ = ws$.pipe(operators_1.filter(isOrderEvent));
    const events$ = Rx.merge(snapshot$, messages$);
    return events$.pipe(operators_1.map(R.cond([
        [isSnapshotEvent, data => normalizeSnapshotEvent(options, data)],
        [isNewOrderEvent, data => normalizeNewOrderEvent(options, data.event)],
        [isFillOrderEvent, data => normalizeFillOrderEvent(options, data.event)],
        [
            isCancelOrderEvent,
            data => normalizeCancelOrderEvent(options, data.event),
        ],
        [
            isRemoveOrderEvent,
            data => normalizeRemoveOrderEvent(options, data.event),
        ],
    ])), operators_1.tap(event => debug.log('Source event: %e', event)), cleanEvents_1.cleanEvents(), operators_1.tap(event => debug.log('Output event: %e', event)));
};
